# -*- coding: utf-8 -*-
#
# Flask-WaffleConf - https://github.com/rmed/flask-waffleconf
#
# Copyright (C) 2015, 2016  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import base64
import pickle

def deserialize(data):
    """Deserialize data using base64 encoding and pickle.

    Arguments:
        data (str): Data to deserialize (must be in base64 and pickled)

    Returns:
        Unpickled object.
    """
    return pickle.loads(base64.b64decode(data.encode()))

def serialize(data):
    """Serialize data using pickle and converting it to a base64 string.

    Arguments:
        data: data to serialize (must be picklable)

    Returns:
        Serialized object.
    """
    return base64.b64encode(pickle.dumps(data)).decode('utf-8')



# -*- coding: utf-8 -*-
#
# Flask-WaffleConf - https://github.com/rmed/flask-waffleconf
#
# Copyright (C) 2015, 2016  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import time

try:
    import redis
    _HAS_REDIS = True

except ImportError:
    _HAS_REDIS = False


def get_watcher(watcher_type):
    """Obtain a watcher function.

    These functions should be executed in a separate thread.

    Arguments:
        watcher_type (str): Either 'file' or 'redis'. If redis is not available,
            it will default to file watcher.

    Returns:
        Watcher function.
    """
    if watcher_type == 'redis' and _HAS_REDIS:
        return _redis_watcher

    else:
        return _file_watcher

def _file_watcher(state):
    """Watch for file changes and reload config when needed.

    Arguments:
        state (_WaffleState): Object that contains reference to app and its
            configstore.
    """
    conf = state.app.config

    file_path = conf.get('WAFFLE_WATCHER_FILE', '/tmp/waffleconf.txt')

    if not os.path.isfile(file_path):
        # Create watch file
        open(file_path, 'a').close()

    while True:
        tstamp = os.path.getmtime(file_path)

        # Compare timestamps and update config if needed
        if tstamp > state._tstamp:
            state.update_conf()
            state._tstamp = tstamp

        # Not too critical
        time.sleep(10)

def _redis_watcher(state):
    """Listen to redis channel for a configuration update notifications.

    Arguments:
        state (_WaffleState): Object that contains reference to app and its
            configstore.
    """
    conf = state.app.config

    r = redis.client.StrictRedis(
        host=conf.get('WAFFLE_REDIS_HOST', 'localhost'),
        port=conf.get('WAFFLE_REDIS_PORT', 6379))

    sub = r.pubsub(ignore_subscribe_messages=True)
    sub.subscribe(conf.get('WAFFLE_REDIS_CHANNEL', 'waffleconf'))

    while True:
        for msg in sub.listen():
            # Skip non-messages
            if not msg['type'] == 'message':
                continue

            tstamp = float(msg['data'])

            # Compare timestamps and update config if needed
            if tstamp > state._tstamp:
                state.update_conf()
                state._tstamp = tstamp

def get_notifier(notifier_type):
    """Obtain a notifier function.

    Arguments:
        notifier_type (str): Either 'file' or 'redis'. If redis is not available,
            it will default to file watcher.

    Returns:
        Notifier function.
    """
    if notifier_type == 'redis' and _HAS_REDIS:
        return _redis_notifier

    else:
        return _file_notifier

def _file_notifier(state):
    """Notify of configuration update through file.

    Arguments:
        state (_WaffleState): Object that contains reference to app and its
            configstore.
    """
    tstamp = time.time()
    state._tstamp = tstamp
    conf = state.app.config

    file_path = conf.get('WAFFLE_WATCHER_FILE', '/tmp/waffleconf.txt')

    if not os.path.isfile(file_path):
        # Create watch file
        open(file_path, 'a').close()

    # Update timestamp
    os.utime(file_path, (tstamp, tstamp))

def _redis_notifier(state):
    """Notify of configuration update through redis.

    Arguments:
        state (_WaffleState): Object that contains reference to app and its
            configstore.
    """
    tstamp = time.time()
    state._tstamp = tstamp
    conf = state.app.config

    # Notify timestamp
    r = redis.client.StrictRedis()
    r.publish(conf.get('WAFFLE_REDIS_CHANNEL', 'waffleconf'), tstamp)

def _dummy(state):
    """Does nothing."""
    pass



# -*- coding: utf-8 -*-
#
# Flask-WaffleConf - https://github.com/rmed/flask-waffleconf
#
# Copyright (C) 2015, 2016  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import

from . import util
from . import watcher
import threading
import time


class _WaffleState(object):
    """Store configstore for the app state.

    This object will update its application's configuration and if the
    ``WAFFLE_MULTIPROC`` setting is set to True, it will also notify other
    processes using Redis channel or file timestamp.

    Arguments:
        app: Flask application instance.
        configstore (WaffleStore): database store.
    """

    def __init__(self, app, configstore):
        self.app = app
        self.configstore = configstore

        # Setup multiprocess notifications
        if self.app.config.get('WAFFLE_MULTIPROC', False):
            op_type = self.app.config.get('WAFFLE_WATCHTYPE', 'file')

            self.watch = watcher.get_watcher(op_type)
            self.notify = watcher.get_notifier(op_type)

            self._tstamp = time.time()

            self._watcher = threading.Thread(target=self.watch, args=(self,))
            self._watcher.setDaemon(True)
            self._watcher.start()

    def parse_conf(self, keys=[]):
        """Parse configuration values from the database.

        The extension must have been previously initialized.

        If a key is not found in the database, it will be created with the
        default value specified.

        Arguments:
            keys (list[str]): list of keys to parse. If the list is empty, then
                all the keys known to the application will be used.

        Returns:
            dict of the parsed config values.
        """
        confs = self.app.config.get('WAFFLE_CONFS', {})
        if not keys:
            keys = confs.keys()

        result = {}

        for key in keys:
            # Some things cannot be changed...
            if key.startswith('WAFFLE_'):
                continue

            # No arbitrary keys
            if key not in confs.keys():
                continue

            stored_conf = self.configstore.get(key)

            if not stored_conf:
                # Store new record in database
                value = confs[key].get('default', '')
                stored_conf = self.configstore.put(key, util.serialize(value))
                self.configstore.commit()

            else:
                # Get stored value
                value = util.deserialize(stored_conf.get_value())

            result[stored_conf.get_key()] = value

        return result

    def update_db(self, new_values):
        """Update database values and application configuration.

        The provided keys must be defined in the ``WAFFLE_CONFS`` setting.

        Arguments:
            new_values (dict): dict of configuration variables and their values
                The dict has the following structure:

                {
                    'MY_CONFIG_VAR'  : <CONFIG_VAL>,
                    'MY_CONFIG_VAR1' : <CONFIG_VAL1>
                }
        """
        confs = self.app.config.get('WAFFLE_CONFS', {})
        to_update = {}

        for key in new_values.keys():
            # Some things cannot be changed...
            if key.startswith('WAFFLE_'):
                continue

            # No arbitrary keys
            if key not in confs.keys():
                continue

            value = new_values[key]
            self.configstore.put(key, util.serialize(value))
            self.configstore.commit()

            to_update[key] = value

        # Update config
        if not to_update:
            return

        self.app.config.update(to_update)

        # Notify other processes
        if self.app.config.get('WAFFLE_MULTIPROC', False):
            self.notify(self)

    def update_conf(self):
        """Update configuration values from database.

        This method should be called when there is an update notification.
        """
        parsed = self.parse_conf()

        if not parsed:
            return None

        # Update app config
        self.app.config.update(parsed)


class WaffleConf(object):
    """Initialize the Flask-WaffleConf extension

    Arguments:
        app: Flask application instance
        configstore (WaffleStore): database store.

    Attributes:
        state (_WaffleState): Direct access to the initialized state object.
    """

    def __init__(self, app=None, configstore=None):
        self.state = None

        if app and configstore:
            self.init_app(app, configstore)

    def init_app(self, app, configstore):
        """Initialize the extension for the given application and store.

        Parse the configuration values stored in the database obtained from
        the ``WAFFLE_CONFS`` value of the configuration.

        Arguments:
            app: Flask application instance
            configstore (WaffleStore): database store.

        """
        if not hasattr(app, 'extensions'):
            app.extensions = {}

        self.state = _WaffleState(app, configstore)
        app.extensions['waffleconf'] = self.state



# -*- coding: utf-8 -*-
#
# Flask-WaffleConf - https://github.com/rmed/flask-waffleconf
#
# Copyright (C) 2015, 2016  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


class WaffleStore(object):
    """Object for connecting to the application database.

    Offers common methods that have to be overriden depending on the
    database type / ORM used.

    Arguments:
        db: Database instance.
        model: Model to work with.
    """

    def __init__(self, db=None, model=None):
        self.db = db
        self.model = model

    def commit(self):
        """Commit to database where needed."""
        raise NotImplementedError

    def delete(self, key):
        """Remove a configuration variable from the database.

        Arguments:
            key (str): Name of the configuration variable to delete.

        Returns:
            Deleted record or ``None`` if it could not be deleted.
        """
        raise NotImplementedError

    def get(self, key):
        """Obtain a configuration variable from the database.

        Arguments:
            key (str): Name of the configuration variable to obtain.

        Returns:
            Record or ``None`` if record could not be obtained.
        """
        raise NotImplementedError

    def put(self, key, value):
        """Insert / Update a configuration variable in the database.

        Arguments:
            key (str): Name of the configuration variable that is being updated.
            value: Value to store in the database (serialized).

        Returns:
            Updated record or ``None`` on error.
        """
        raise NotImplementedError


class AlchemyWaffleStore(WaffleStore):
    """Config store for SQLAlchemy."""

    def commit(self):
        self.db.session.commit()

    def delete(self, key):
        record = self.model.query.filter_by(key=key).first()

        if not record:
            return None

        self.db.session.delete(record)

        return record

    def get(self, key):
        return self.model.query.filter_by(key=key).first()

    def put(self, key, value):
        record = self.model.query.filter_by(key=key).first()

        if not record:
            # Creating new record
            record = self.model()
            record.key = key
            record.value = value

            self.db.session.add(record)

        else:
            # Updating record
            record.value = value

        return record


class PeeweeWaffleStore(WaffleStore):
    """Config store for peewee."""

    def commit(self):
        if self.db.get_autocommit():
            self.db.commit()

    def delete(self, key):
        record = self.get(self.model, key)

        try:
            record.delete_instance()
            return record

        except:
            return None

    def get(self, key):
        try:
            return self.model.get(self.model.key == key)

        except self.model.DoesNotExist:
            # Does not exist
            return None

    def put(self, key, value):
        record = self.get(key)

        if not record:
            # Creating new record
            record = self.model()
            record.key = key

        record.value = value

        record.save()

        return record



# -*- coding: utf-8 -*-
#
# Flask-WaffleConf - https://github.com/rmed/flask-waffleconf
#
# Copyright (C) 2015  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from setuptools import setup
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description
with open(path.join(here, 'DESCRIPTION.rst')) as f:
    long_description = f.read()

setup(
    name='Flask-WaffleConf',
    version='0.3.1',
    url='https://github.com/rmed/flask-waffleconf',
    license='GPLv2+',
    author='Rafael Medina García',
    author_email='rafamedgar@gmail.com.com',
    description='Store variables in database and update them at runtime',
    long_description=long_description,
    packages=['flask_waffleconf'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'Flask'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# -*- coding: utf-8 -*-
#
# Flask-WaffleConf - https://github.com/rmed/flask-waffleconf
#
# Copyright (C) 2015, 2016  Rafael Medina García <rafamedgar@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


class WaffleMixin(object):
    """Mixin used in the creation of the database model.

    WaffleConf expects a model that has (at least) the following fields:

        - key (str): Unique identifier for configuration variable.
        - value (str): Value for the configuration variable.

    Values may be a large string, so make sure to define a field capable
    of storing big strings. These values are later parsed according to the
    ``type`` specified in the application configuration.
    """

    def get_key(self):
        """Obtain the key for the configuration variable.

        Mixin expects a ``self.key`` attribute (str) in the model.
        If this is not the case, you should override this method.

        Returns:
            Configuration key string.
        """
        try:
            return self.key

        except AttributeError:
            raise NotImplementedError("Model does not have attribute 'key'")

    def get_value(self):
        """Obtain the value for the configuration variable.

        Mixin expects a ``self.value`` attribute (str) in the model.
        If this is not the case, you should override this method.

        Returns:
            Configuration value string.
        """
        try:
            return self.value

        except AttributeError:
            raise NotImplementedError("Model does not have attribute 'value'")
